package buckets

import (
	"errors"

	"git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-lens/internal/schema/common"
	cid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/user"
	"github.com/mr-tron/base58"
)

type (
	PrefixBucket struct {
		prefix    Prefix
		resolvers Resolvers
	}

	PrefixContainerBucket struct {
		prefix    Prefix
		id        cid.ID
		resolvers Resolvers
	}

	ContainerBucket struct {
		id        cid.ID
		resolvers Resolvers
	}

	UserBucket struct {
		id        user.ID
		resolvers Resolvers
	}

	UserAttributeKeyBucket struct {
		prefix Prefix
		id     cid.ID
		key    string
	}

	UserAttributeValueBucket struct {
		value string
	}
)

type (
	FilterResolver = func(result bool) common.FilterResult

	Resolvers struct {
		cidResolver FilterResolver
		oidResolver FilterResolver
	}
)

var (
	StrictResolver  = func(x bool) common.FilterResult { return common.IfThenElse(x, common.Yes, common.No) }
	LenientResolver = func(x bool) common.FilterResult { return common.IfThenElse(x, common.Yes, common.Maybe) }
)

var (
	ErrNotBucket          = errors.New("not a bucket")
	ErrInvalidKeyLength   = errors.New("invalid key length")
	ErrInvalidValueLength = errors.New("invalid value length")
	ErrInvalidPrefix      = errors.New("invalid prefix")
)

func NewPrefixBucketParser(prefix Prefix, next common.Parser, resolvers Resolvers) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		if len(key) != 1 {
			return nil, nil, ErrInvalidKeyLength
		}
		var b PrefixBucket
		if b.prefix = Prefix(key[0]); b.prefix != prefix {
			return nil, nil, ErrInvalidPrefix
		}
		b.resolvers = resolvers
		return &b, next, nil
	}
}

func NewPrefixContainerBucketParser(prefix Prefix, next common.Parser, resolvers Resolvers) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		if len(key) != 33 {
			return nil, nil, ErrInvalidKeyLength
		}
		var b PrefixContainerBucket
		if b.prefix = Prefix(key[0]); b.prefix != prefix {
			return nil, nil, ErrInvalidPrefix
		}
		if err := b.id.Decode(key[1:]); err != nil {
			return nil, nil, err
		}
		b.resolvers = resolvers
		return &b, next, nil
	}
}

func NewUserBucketParser(next common.Parser, resolvers Resolvers) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		var b UserBucket
		if err := b.id.DecodeString(base58.Encode(key)); err != nil {
			return nil, nil, err
		}
		b.resolvers = resolvers
		return &b, next, nil
	}
}

func NewContainerBucketParser(next common.Parser, resolvers Resolvers) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		if len(key) != 32 {
			return nil, nil, ErrInvalidKeyLength
		}
		var b ContainerBucket
		if err := b.id.Decode(key); err != nil {
			return nil, nil, err
		}
		b.resolvers = resolvers
		return &b, next, nil
	}
}

func NewUserAttributeKeyBucketParser(next common.Parser) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		if len(key) < 34 {
			return nil, nil, ErrInvalidKeyLength
		}
		var b UserAttributeKeyBucket
		if b.prefix = Prefix(key[0]); b.prefix != UserAttribute {
			return nil, nil, ErrInvalidPrefix
		}
		if err := b.id.Decode(key[1:33]); err != nil {
			return nil, nil, err
		}
		b.key = string(key[33:])
		return &b, next, nil
	}
}

func NewUserAttributeValueBucketParser(next common.Parser) common.Parser {
	return func(key, value []byte) (common.SchemaEntry, common.Parser, error) {
		if value != nil {
			return nil, nil, ErrNotBucket
		}
		if len(key) == 0 {
			return nil, nil, ErrInvalidKeyLength
		}
		var b UserAttributeValueBucket
		b.value = string(key)
		return &b, next, nil
	}
}
